import threading
from . import specs
try: from urllib.parse import quote_plus
except ImportError: from urllib import quote_plus	
import skitai
from skitai.saddle import part

class WebSocketServers:
	def __init__ (self, wasc):
		self.wasc = wasc
		self._nomore = False
		self.lock = threading.RLock ()
		self.wss = {}
	
	def get (self, gid, default = None):
		with self.lock:
			return self.wss.get (gid, default)		
		
	def has_key (self, gid):	
		with self.lock:
			has = gid in self.wss		
		return has
		
	def create (self, gid, *args):	
		with self.lock:
			if self._nomore: 
				return
				
		wss = WebSocketServer (gid, *args)
		with self.lock:
			self.wss [gid] = wss		
		return wss
		
	def remove (self, gid):
		with self.lock:
			try: 
				del self.wss [gid]			
			except KeyError: 
				pass		
	
	def close (self):
		with self.lock:
			wss = list (self.wss.items ())
			self._nomore = True
		for k, s in wss:
			self.wasc.logger ('server', '...closing websockket %s' % k)
			s.close ()
	
	def cleanup (self):
		self.close ()
		

class WebSocketServer (specs.WebSocket1):
	def __init__ (self, gid, handler, request, apph, env, message_encoding = None):
		specs.WebSocket.__init__ (self, handler, request, message_encoding)
		self._closed = False
		self.gid = gid
		self.apph = apph
		self.env = env
		self.messages = []
		self.clients = {}		
			
	def add_client (self, ws):
		self.clients [ws.client_id] = ws
		ws.handle_message (1, skitai.WS_EVT_OPEN)
	
	def make_params (self, msg, event):
		specs.WebSocket1.make_params (self, msg, event)
		
	def handle_message (self, client_id, msg, querystring, params, message_param):		
		if msg == -1: # exit
			try: del self.clients [client_id]
			except KeyError: pass
			msg = ""			
			if not self.clients:
				return self.close ()
		elif msg == 1: # enter
			msg = ""
			
		self.env ["QUERY_STRING"] = querystring + quote_plus (msg)			
		self.env ["websocket.params"] = params
		self.env ["websocket.params"][message_param] = self.message_decode (msg)		
		args = (self.request, self.apph, (self.env, self.start_response), self.wasc.logger)
		if self.env ["wsgi.multithread"]:
			self.wasc.queue.put (specs.PooledJob (*args))
		else:
			specs.PooledJob (*args) ()
		
	def close (self):
		websocket_servers.remove (self.gid)
		self.clients = {}
		self._closed = True
			
	def send (self, msg, client_id = None, op_code = -1):
		msg, op_code = self.build_data (msg, op_code)						
		if client_id:
			self.__sendto (client_id, msg, op_code)
		else:
			self.__sendall (msg, op_code)
	
	def __sendall (self, msg, op_code = -1):
		clients = list (self.clients.keys ())
		for client_id in clients:
			self.__sendto (client_id, msg, op_code)
			
	def __sendto (self, client_id, msg, op_code = -1):
		try:
			client = self.clients [client_id]
		except KeyError:
			client = None	
		if client:
			client.send (msg, op_code)
		
	
websocket_servers = None

def start_websocket (wasc):
	global websocket_servers
	
	websocket_servers = WebSocketServers (wasc)
	
	



import base64
from hashlib import md5

class Authorizer:
	def __init__ (self, realm, apikey, password):
		self.realm = realm
		self.apikey = apikey.encode ("utf8")
		self.md5_apikey = md5 (self.apikey).hexdigest ()
		self.password = password
		self.cache = {}
	
	def get_authority (self, authkey, request = None):
		authinfo = request.get_header (authkey)		
		if not authinfo: return None
		
		if authinfo in self.cache:
			return self.cache [authinfo]
		
		basic = base64.decodestring (authinfo [6:])
		if basic == self.md5_apikey:
			self.cache [authinfo] = "app"
			return "app"
		
		username, password = basic.split (":")				
		if username != "admin": 
			return None
		
		try:
			assert (password == self.password)
		except AssertionError:
			return None
		else:
			self.cache [authinfo] = "admin"
			return "admin"
	
	def _has_permission (self, group, permission):
		if not permission:
			return True #all
		if group == "admin":
			return True
		else:
			return group in permission
					
	def has_permission (self, request, permission = None, push_error = 1):
		group = self.get_authority ('Authorization', request)
		if group is None:
			request.response ['WWW-Authenticate'] = 'Basic realm="%s Secured"' % self.realm
			if push_error:
				request.response.error (401)
			return False
		if not self._has_permission (group, permission):
			if push_error: 
				request.response.error (403, "", "You haven't permission for accessing this page")
			return False
		return True


class FileAuthorizer (Authorizer):
	cache = {}
	def __init__ (self, realm, apikey):
		self.realm = realm
		self.apikey = apikey
		self.md5_apikey = md5.new (self.apikey).hexdigest ()
		self.userfile = None
		self.users = {}		
		
	def register (self, username, image, group):
		self.users [username] = (image, group)		
	
	def load (self, userfile):
		f = open (userfile, 'r')
		for line in f:
			line = line.strip ()
			if not line: break
			username, group, image = line.split (":")				
			self.register (username, image, group)						
		f.close ()
		self.userfile = userfile
		
	def digest (self, username, group, password):
		return md5.new ("%s:%s:%s" % (username, group, password)).hexdigest ()
		
	def useradd (self, username, password, group):
		self.register (username, self.digest (username, group, password), group)
		self.save ()
		return True
		
	def usermod (self, username, password, group):
		self.useradd (username, password, group)
		self.save ()
		return True
		
	def userdel (self, username):
		try:
			del self.users [username]
		except KeyError:
			return True	
		self.save ()
		return True
	
	def isvaliduser (self, username, password):
		return self.digest (username, password) == self.users [username][0]
			
	def save (self, fn = None):
		f = open (self.userfile, "w")
		for username, (image, grp) in list(self.users.items ()):
			f.write ("%s:%s:%s\n" % (username, grp, image))
		f.close ()
			
	def get_authority (self, authkey, request = None):
		authinfo = request.get_header (authkey)		
		if not authinfo: return None
		
		if authinfo in self.cache:
			return self.cache [authinfo]
		
		basic = base64.decodestring (authinfo [6:])
		if basic == self.md5_apikey:
			return "app"
		
		username, password = basic.split (":")				
		try:
			group = self.users [username][1]
		except KeyError:
			return None

		image = self.digest (username, group, password)		
		try:
			assert (image == self.users [username][0])
		except AssertionError:
			return None
		else:
			self.cache [authinfo] = group
			return group
	
	def get_userinfo (self, authkey, request = None):	
		authinfo = request.get_header (authkey)		
		if not authinfo: return None
		
		basic = base64.decodestring (authinfo [6:])
		if basic == self.md5_apikey:
			return None, "app"
		
		username, password = basic.split (":")				
		group = self.users [username][1]
		return username, group
		
		